from argparse import ArgumentParser, RawDescriptionHelpFormatter
import hashlib
import json
from pathlib import Path
import pickle
import os
import shlex
import subprocess
from io import StringIO
from tqdm import tqdm
import pandas as pd
from typing import Mapping
import plotly.express as px

from destine_analytics.lsreport import read_filetree
from .utils import sanitize_output


def _get_parser():
    parser = ArgumentParser(
        description="Parser for output of `tools/create_filetree.sh`.",
        formatter_class=RawDescriptionHelpFormatter,
    )
    parser.add_argument(
        "path",
        type=str,
        help=(
            "If it is a directory, local (or remote) path to draw filetree from. "
            "If it is a file, it is assumed to be the output of tools/create_filetree.sh."
        ),
    )
    parser.add_argument(
        "-e",
        "--ssh-command",
        type=str,
        help=(
            "For remote path scoping: ssh command to execute in order to access to remote shell."
        ),
    )
    parser.add_argument(
        "-E",
        "--expid",
        type=str,
        default=None,
        help="Expid. Only for title & filename purposes",
    )
    parser.add_argument(
        "-r",
        "--root",
        type=str,
        default=None,
        help="Root directory to display. Defaults to the main path.",
    )
    parser.add_argument(
        "-d",
        "--only-directories",
        action="store_true",
        help="Only show directories in the figure. Needed for excessively large file trees.",
    )
    parser.add_argument(
        "-f",
        "--figure",
        type=str,
        default=None,
        choices=["treemap", "sunburst", "icicle", "all", None],
        help="Figure type to generate.",
    )
    parser.add_argument(
        "-i",
        "--interactive",
        action="store_true",
        help="Whether to interactively show the figure(s) after plotting.",
    )
    parser.add_argument(
        "-c",
        "--cache",
        action="store_true",
        help="Read / write to ~/.cache",
    )
    parser.add_argument(
        "-o",
        "--output",
        nargs="?",
        type=str,
        default=None,
        const="__default__",
        help=(
            "Output directory or file for saving figures. If no path is given, "
            "defaults to 'figures/'. If flag is not used, figure is not saved."
        ),
    )
    return parser


def _json_serialize(obj):
    if isinstance(obj, Path):
        return str(obj)
    raise TypeError(f"Object of type {type(obj).__name__} is not JSON serializable")


def _get_cache_path(name, *args, **kwargs) -> Path:
    key = json.dumps([name, *args, kwargs], default=_json_serialize)
    filename = hashlib.sha256(key.encode("utf-8")).hexdigest()
    directory = os.path.basename(__file__)
    return Path.home().resolve() / ".cache" / directory / filename


def _cache_read(cache, name, /, *args, **kwargs):
    if cache is False:
        return
    cpath = _get_cache_path(name, *args, **kwargs)
    if not cpath.exists():
        return None
    print(f"Reading '{name}' from cache...")
    with open(cpath, "rb") as fb:
        return pickle.load(fb)


def _cache_write(cache, obj, name, /, *args, **kwargs):
    if cache is False:
        return
    print(f"Writing '{name}' to cache...")
    cpath = _get_cache_path(name, *args, **kwargs)
    if not cpath.parent.exists():
        cpath.parent.mkdir(parents=True)
    with open(cpath, "wb") as fb:
        pickle.dump(obj, fb)


def _add_path_column(df):
    # For create_filetree.sh output, Parent already contains the full directory path
    # We don't need to normalize it further, just create the full file paths
    df["Path"] = df.apply(lambda row: os.path.join(row["Parent"], row["Filename"]), axis=1)
    return df


def _get_subtree_df(df, root):
    # For create_filetree.sh output, Parent contains full directory paths
    df = df.copy()

    # Normalize requested root
    if root in (None, ""):
        # No root specified, return all data
        subtree_df = df.copy()
    else:
        # If root is absolute, use it directly
        if os.path.isabs(root):
            abs_root = os.path.normpath(root)
        else:
            # If root is relative, we need to find the common root first
            common_root = os.path.commonpath(list(df["Parent"]))
            abs_root = os.path.normpath(os.path.join(common_root, root))

        # Filter to subtree - only include files that are under the absolute root
        if abs_root == "/":
            mask = df["Parent"].str.startswith("/")
        else:
            mask = (df["Parent"] == abs_root) | (
                df["Parent"].str.startswith(abs_root + "/")
            )
        subtree_df = df[mask].copy()

        if subtree_df.empty:
            raise ValueError(
                f"No entries found under root '{root}' (resolved to '{abs_root}')."
            )

    subtree_df["name"] = subtree_df["Filename"]

    return subtree_df


def _human_readable_size(size_bytes: int) -> str:
    """Convert size in bytes to human readable format with ~3 significant digits."""
    if size_bytes == 0:
        return "0B"
    for unit in ("B", "K", "M", "G", "T", "P", "E", "Z"):
        if abs(size_bytes) < 1024.0:
            if unit == "B":
                return f"{size_bytes:.0f}{unit}"
            if size_bytes < 10:
                return f"{size_bytes:.2f}{unit}"
            if size_bytes < 100:
                return f"{size_bytes:.1f}{unit}"
            return f"{size_bytes:.0f}{unit}"
        size_bytes /= 1024.0
    return f"{size_bytes:.2f}Y"


def draw(
    sizes: Mapping[str, int],
    parents: Mapping[str, str],
    root: str = None,
    figure: str = "treemap",
    expid: str = None,
):
    """
    Visualize file structure under a given root directory using Plotly.

    Parameters:
        sizes (Mapping[str, int]): Dictionary mapping directory paths to their total sizes
        parents (Mapping[str, str]): Dictionary mapping directory paths to their parent directories
        root (str | None): Subdirectory to start visualization from. Can be absolute or relative to the common root.
        figure (str): One of 'treemap', 'sunburst', or 'icicle'.
    """

    if figure not in {"treemap", "sunburst", "icicle"}:
        raise ValueError(
            f"Unsupported figure '{figure}'. Use 'treemap', 'sunburst', or 'icicle'."
        )

    # Create visualization DataFrame
    viz_df = pd.DataFrame(
        {
            "id": list(sizes.keys()),  # Full path as ID
            "parent": list(parents.values()),  # Full parent path
            "value": list(sizes.values()),  # Directory size
        }
    )

    # Add display name (last part of path + size)
    viz_df["name"] = viz_df.apply(
        lambda row: f"{row['id'].split('/')[-1]} ({_human_readable_size(row['value'])})",
        axis=1,
    )

    # Add human readable size for hover
    viz_df["size_str"] = viz_df["value"].apply(_human_readable_size)

    # If a root is specified, filter the data to start from there.
    if root:
        # This is safer than `startswith` which can match `/path/to/dir-other`
        if root == "/":
            mask = viz_df["id"].str.startswith("/")
        else:
            mask = (viz_df["id"] == root) | viz_df["id"].str.startswith(root + "/")
        viz_df = viz_df[mask]

        if viz_df.empty:
            raise ValueError(f"No entries found under root '{root}'.")

    # Any node whose parent is not in the dataset is a root.
    all_ids = set(viz_df["id"])
    viz_df.loc[~viz_df["parent"].isin(all_ids), "parent"] = ""

    # Create the figure
    fig = getattr(px, figure)(
        viz_df,
        ids="id",  # Use full paths as IDs
        names="name",  # Show name with size
        parents="parent",  # Use full parent paths
        values="value",
        hover_data=[
            "id",
            "size_str",
        ],  # Show full path and human readable size in hover
    )

    # Get the root for the title
    root_entries = viz_df[viz_df["parent"] == ""]
    if not root_entries.empty:
        abs_root = root_entries["id"].iloc[0]
    elif not viz_df.empty:
        abs_root = viz_df["id"].iloc[0]
    else:
        abs_root = root or "unknown"

    # Update traces for better visibility
    fig.update_traces(
        root_color="lightgrey",
        textposition="middle center",  # Center text in each box
        textfont=dict(size=14),  # Larger text
        textinfo="label",  # Show the label (name) in each box
    )

    # Update layout for better visibility and larger size
    title_text = f"{figure.title()} of directory sizes under: {abs_root}"
    if expid:
        title_text = f"[{expid}] {title_text}"
    fig.update_layout(
        title=dict(
            text=title_text,
            font=dict(size=24),  # Larger title
        ),
        margin=dict(t=60, l=0, r=0, b=0),  # More space for title
        width=2000,  # Wider figure
        height=1500,  # Taller figure
        font=dict(size=14),  # Larger font for all text
    )

    return fig


def makefigure(
    sizes: Mapping[str, int],
    parents: Mapping[str, str],
    root: str,
    figure: str,
    interactive: bool,
    output_dir: str,
    prefix: str,
    cache: bool,
    **kwargs,
):
    if figure == "all":
        return [
            makefigure(
                sizes,
                parents,
                root,
                "treemap",
                interactive,
                output_dir,
                prefix,
                cache,
                **kwargs,
            ),
            makefigure(
                sizes,
                parents,
                root,
                "sunburst",
                interactive,
                output_dir,
                prefix,
                cache,
                **kwargs,
            ),
            makefigure(
                sizes,
                parents,
                root,
                "icicle",
                interactive,
                output_dir,
                prefix,
                cache,
                **kwargs,
            ),
        ]

    print(f"Plotting {figure}.")
    fig = _cache_read(cache, "figure", root=root, figure=figure, **kwargs)
    if fig is None:
        fig = draw(sizes, parents, root, figure, kwargs.get("expid"))
        _cache_write(cache, fig, "figure", root=root, figure=figure, **kwargs)

    filename = f"{prefix}{figure}.png"
    save_path = sanitize_output(output_dir, filename)
    if save_path:
        save_path.parent.mkdir(parents=True, exist_ok=True)
        fig.write_image(save_path)
    if interactive:
        print("Showing figure...")
        fig.show()

    return fig


def compute_directory_size(df, only_directories=False):
    """
    Compute directory sizes and their hierarchical relationships.

    Parameters:
        df (pd.DataFrame): DataFrame with 'Parent' and 'Filename' columns
        only_directories (bool): If True, only show directories in the visualization

    Returns:
        tuple: (sizes, parents) where:
            - sizes: dict mapping paths to their total sizes
            - parents: dict mapping paths to their parent directories
    """
    if df.empty:
        return {}, {}

    total_sizes = {}
    parents = {}

    # Step 1: Process all files, recording their size and parent.
    for _, row in tqdm(df.iterrows(), total=len(df), desc="Processing files"):
        parent_path = row["Parent"]
        file_path = os.path.join(parent_path, row["Filename"])
        total_sizes[file_path] = int(row["Filesize"])
        parents[file_path] = parent_path

    # Step 2: Ensure all directories and their ancestors are in the data structures.
    all_paths = set(parents.keys())
    for path in all_paths:
        parent = parents.get(path)
        while parent and parent not in parents:
            if parent not in total_sizes:
                total_sizes[parent] = 0

            grandparent = os.path.dirname(parent)
            if grandparent == parent:  # Reached the root
                parents[parent] = ""
                break
            parents[parent] = grandparent
            parent = grandparent

    # Step 3: Propagate sizes up the tree from the longest paths to the shortest.
    for path in sorted(total_sizes.keys(), key=len, reverse=True):
        parent = parents.get(path)
        if parent is not None and parent in total_sizes:
            total_sizes[parent] += total_sizes[path]

    # If only directories are requested, filter out the files.
    if only_directories:
        # A directory is any path that is a parent of another path.
        dir_paths = set(parents.values())

        # Filter total_sizes and parents to only include directories.
        total_sizes = {path: size for path, size in total_sizes.items() if path in dir_paths}
        parents = {path: parent for path, parent in parents.items() if path in dir_paths}

    # Any node whose parent is not in the dataset is a root.
    all_paths = set(total_sizes.keys())
    for path, parent in parents.items():
        if parent not in all_paths:
            parents[path] = ""

    return total_sizes, parents


def _get_dataframe(path_str, ssh_command, expid, cache):
    if ssh_command:
        find_cmd = f"find {shlex.quote(path_str)} -ls"
        ssh_cmd_list = shlex.split(ssh_command) + [find_cmd]
        try:
            process = subprocess.run(
                ssh_cmd_list,
                capture_output=True,
                text=True,
                check=True,
                encoding="utf-8",
            )
            return read_filetree(StringIO(process.stdout), common=False)
        except (subprocess.CalledProcessError, FileNotFoundError) as e:
            raise RuntimeError(f"Error executing remote command: {e}") from e
    else:
        path = Path(path_str)
        if not path.exists():
            raise FileNotFoundError(f"Local path '{path}' does not exist.")

        if path.is_dir():
            print(f"Scanning directory '{path}'...")
            try:
                process = subprocess.run(
                    ["find", str(path), "-ls"],
                    capture_output=True,
                    text=True,
                    check=True,
                    encoding="utf-8",
                )
                return read_filetree(StringIO(process.stdout), common=False)
            except subprocess.CalledProcessError as e:
                raise FileNotFoundError(
                    f"Error executing 'find' on '{path}': {e}"
                ) from e
        if path.is_file():
            return read_filetree(path, common=False)
        raise FileNotFoundError(f"Local path '{path}' is not a directory or a file.")


def main():

    args = _get_parser().parse_args()
    path_str = os.path.normpath(args.path)
    ssh_command = args.ssh_command
    expid = args.expid
    root = os.path.normpath(args.root) if args.root is not None else path_str
    only_directories = args.only_directories
    figure = args.figure
    interactive = args.interactive
    cache = args.cache
    output_dir = args.output
    df = None

    cache_key_args = {
        "path": path_str,
        "expid": expid,
        "ssh_command": ssh_command,
    }
    df = _cache_read(cache, "df", **cache_key_args)
    if df is None:
        df = _get_dataframe(path_str, ssh_command, expid, cache)
        _cache_write(cache, df, "df", **cache_key_args)

    df["Parent"] = df["Parent"].apply(os.path.normpath)
    df = _add_path_column(df)
    df = _get_subtree_df(df, root)

    kwargs = {
        "path": path_str,
        "expid": expid,
        "root": root,
        "only_directories": only_directories,
        "ssh_command": ssh_command,
    }
    sizes_parents = _cache_read(cache, "sizes+parents", **kwargs)
    if sizes_parents is None:
        sizes, parents = compute_directory_size(df, only_directories=only_directories)
        sizes_parents = (sizes, parents)
        _cache_write(cache, sizes_parents, "sizes+parents", **kwargs)

    if figure is not None:
        if interactive and figure == "all":
            raise ValueError(
                "Interactive mode cannot be used with --figure all. Please select a single figure type."
            )
        makefigure(
            sizes=sizes_parents[0],
            parents=sizes_parents[1],
            figure=figure,
            interactive=interactive,
            output_dir=output_dir,
            prefix=(f"{expid}_" if expid else ""),
            cache=cache,
            **kwargs,
        )


if __name__ == "__main__":
    main()
